import os
import sys
import fcntl
import popen2
import re
import string

def GetCommandOutput(command):
	proc = popen2.Popen4(command)
	output = proc.fromchild.read()
	rc = proc.wait()
	return rc,output

def GetCommandOutput_NoStderr(command):
	proc = popen2.Popen3(command)
	output = proc.fromchild.read()
	rc = proc.wait()
	return rc,output

def FileList(pattern,path="./"):
	list = []
	if len(pattern) and pattern[-1] <> "$":
		pattern = pattern + "$"
	p = re.compile(pattern)
	for f in os.listdir(path):
		if p.match(f):
			list.append(os.path.join(path,f))
	return list

def ReplaceInFilename(fname,replace_tag,replacement):
	if string.find(fname,replace_tag) < 0: return fname

	if type(replacement) == type([]):
		if len(replacement) <> 1:
			sys.stderr.write("Cannot replace '" + str(replace_tag) \
			                 + "' in '" + str(fname) + "' because there are " \
			                 + str(len(replacement)) + " options: " \
			                 + str(replacement) + "\n")
			sys.exit(1)
		replacement = replacement[0]

	return string.replace(fname,replace_tag,replacement)

def split_url(url,default_protocol="file"):
	protocol = default_protocol
	host = None
	port = None
	path = None
	params = None

	pos = string.find(url,":")
	if pos >= 0:
		protocol = url[0:pos]
		pos = pos + 1
		if string.find(url,"//",pos) == pos:
			pos = pos + 2

		url = url[pos:]

		host = ""
		pos = 0
		while pos < len(url) and (
			url[pos] in string.ascii_letters or \
			url[pos] in string.digits or \
			url[pos] == "."):

			host = host + str(url[pos])
			pos = pos + 1
		url = url[pos:]

		if len(url) and url[0] == ":":
			pos = 1
			port = ""
			while pos < len(url) and (url[pos] in string.digits):
				port = port + url[pos]
				pos = pos + 1
			if port:
				port = int(port)
			url = url[pos:]

	pos = string.find(url,"?")
	if pos < 0:
		path = url
	else:
		path = url[0:pos]
		url = url[pos+1:]
		params = url

	return (protocol,host,port,path,params)

def ChooseURL(urls,append_path=None):
	if not urls: return urls
	if type(urls) == type(""):
		if string.find(urls,"|") >= 0:
			#This is the old way of doing it (for backward compatibility).
			urls = string.split(urls,"|")
		else:
			urls = string.split(urls,"\n")


	best_url = None
	for url in urls:
		if not url: continue
		if append_path:
			url = os.path.join(url,append_path)
		best_url = url
		protocol,host,port,path,params = split_url(url)
		if protocol != "file": break #assume this URL exists

		#It's a file url, so verify that it exists.
		if os.path.exists(path):
			#It exists, so stop looking for a better choice.
			break

	return best_url

def PoolCatalogUrlToFile(pool_file_catalog):
	if string.find(pool_file_catalog,"file:") == 0:
		return pool_file_catalog[5:]
	elif string.find(pool_file_catalog,"/") == 0:
		return pool_file_catalog
	else:
		raise "Unsupported POOL Catalog URL type."

def LockPoolForWrite(pool_file_catalog):
	fname = PoolCatalogUrlToFile(pool_file_catalog)
	locked_pool_file = file(fname + ".lock","w+")
	fcntl.lockf(locked_pool_file,fcntl.LOCK_EX)
	return locked_pool_file

def UnlockPool(locked_pool_file):
	fcntl.lockf(locked_pool_file,fcntl.LOCK_UN)
	locked_pool_file.close()

def CopyFile(src,dest):
	command = "cp '" + str(src) + "' '" + str(dest) + "'"
	rc = os.system(command)
	if rc:
		raise "Failed to copy file."

def SingleToList(item):
	if type(item) <> type([]):
		if item:
			return [item]
		else:
			return []
	return item

def GetMetafileList(pool_file_catalog,owners):
	owners = SingleToList(owners)
	files = []
	#For now, assume meta data is in same directory as pool catalog.
	metadata_path = os.path.dirname(PoolCatalogUrlToFile(pool_file_catalog))
	for owner in owners:
		if not owner: continue
		for fname in os.listdir(metadata_path):
			if re.match("""^.*\.META\.""" + owner,fname):
				files.append(os.path.join(metadata_path,fname))
	return files

def ChangeMetaDataPFNs(meta_path,pool_file_catalog,owner):
	if len(meta_path) == 0 or meta_path[-1] <> "/":
		meta_path = meta_path + "/"

	old_fname = PoolCatalogUrlToFile(pool_file_catalog)
	new_fname = old_fname + ".new"
	command = """sed 's|\(<pfn .*name="\).*/\([^/"]*\.META.""" + owner + """\"\)|\\1""" + meta_path + """\\2|g' """ + old_fname + " > " + new_fname
	rc = os.system(command)
	if rc:
		sys.stderr.write("The following command failed:\n" + command + "\n")
		sys.stderr.write("Failed to change meta data path.")
		sys.exit(1)
	command = "mv " + new_fname + " " + old_fname
	rc = os.system(command)
	if rc:
		sys.stderr.write("The following command failed:\n" + command + "\n")
		sys.exit(1)

def CheckoutPool(pool_file_catalog,owners):
	owners = SingleToList(owners)
	files = GetMetafileList(pool_file_catalog,owners)
	pool_fname = PoolCatalogUrlToFile(pool_file_catalog)
	files.append(pool_fname)

	for fname in files:
		dest = os.path.basename(fname)
		if not os.path.exists(dest):
			#META files already exist locally for meta init jobs
			CopyFile(fname,dest)

	local_pool_file_catalog = "file:./" + os.path.basename(pool_fname)
	for owner in owners:
		ChangeMetaDataPFNs("./",local_pool_file_catalog,owner)
	return local_pool_file_catalog

def UpdatePoolLog(pool_file_catalog,message):
	pool_log_path = os.path.dirname(PoolCatalogUrlToFile(pool_file_catalog))
	F = file(os.path.join(pool_log_path,"pool.log"),"a")
	if len(message) == 0 or message[-1] <> "\n":
		message = message + "\n"
	F.write(message)
	F.close()

def CommitPool(local_pool_file_catalog,pool_file_catalog,owners):
	#First copy all of the files into the same directory,
	#then rename them to replace the old files.

	owners = SingleToList(owners)
	mfiles = GetMetafileList(local_pool_file_catalog,owners)
	metadata_path = os.path.dirname(PoolCatalogUrlToFile(pool_file_catalog))
	files = {}
	for fname in mfiles:
		f = os.path.basename(fname)
		files[f] = os.path.join(metadata_path,f)

	files[PoolCatalogUrlToFile(local_pool_file_catalog)] = PoolCatalogUrlToFile(pool_file_catalog)

	for owner in owners:
		ChangeMetaDataPFNs(metadata_path,local_pool_file_catalog,owner)

	for src in files.keys():
		dest = files[src] + ".new"
		CopyFile(src,dest)

	try:
		for src in files.keys():
			dest = files[src]
			src = dest + ".new"
			os.rename(src,dest)
	except:
		sys.stderr.write("ERROR!  Failed in CRITICAL section.\n")
		raise

def GetAttributeFromXMLFiles(files,attribute):
	#For now, just use grep/sed...
	files = string.join(files)
	command = \
	  "grep 'att_name=\"" + attribute + "\"' " + files + \
	  """ | sed 's/.*att_value="\([^"]*\)\".*/\\1/g' | sort | uniq"""
	rc,values = GetCommandOutput(command)
	if rc:
		sys.stderr.write("Failed to execute the following command:\n" + command + "\n")
		sys.exit(1)

	values = string.split(values,'\n')
	while ('' in values):
		values.remove('')

	return values

POOL_HEADER = \
"""<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
<!-- Edited By POOL -->
<!DOCTYPE POOLFILECATALOG SYSTEM "InMemory">
<POOLFILECATALOG>
  <META name="Content" type="string"/>
  <META name="DBoid" type="string"/>
  <META name="DataType" type="string"/>
  <META name="FileCategory" type="string"/>
  <META name="Flags" type="string"/>
  <META name="dataset" type="string"/>
  <META name="jobid" type="string"/>
  <META name="owner" type="string"/>
  <META name="runid" type="string"/>

  <File ID="10986CC7-9F04-D811-983D-00D0B7B8714E">
	<physical>
	  <pfn filetype="fake" name="PersilCatalogInit"/>
	</physical>
	<logical/>
	<metadata att_name="DBoid" att_value=""/>
	<metadata att_name="DataType" att_value=""/>
	<metadata att_name="FileCategory" att_value=""/>
	<metadata att_name="dataset" att_value=""/>
	<metadata att_name="jobid" att_value=""/>
	<metadata att_name="owner" att_value=""/>
	<metadata att_name="runid" att_value=""/>
  </File>
"""
POOL_FOOTER = """
</POOLFILECATALOG>
"""

def Get_EVD_PFN(name):
	#Look up the URL for the given file.
	#This should have been stored in a file of the same name
	#in the working directory.

	F = open(name,"r")
	url = F.read()
	F.close()
	return url

def Replace_EVD_PFN(m):
	header = m.group(1)
	orig_path = m.group(2)
	trailer = m.group(3)

	new_path = Get_EVD_PFN(os.path.basename(orig_path))

	return header + new_path + trailer

def GetFinalPoolXML(initial_xml):
	return re.sub(
	  """(<pfn .*name=\")([^\"]*)(\".*>)""",
	  Replace_EVD_PFN,
	  initial_xml)

def MakeEVDCatalog(xml_fragment_files,tmp_evd_catalog):
	F = open(tmp_evd_catalog,"w")
	F.write(POOL_HEADER)
	for fname in xml_fragment_files:
		initial_xml = file(fname).read()
		final_xml = GetFinalPoolXML(initial_xml)
		F.write(final_xml)
	F.write(POOL_FOOTER)
	F.close()

def ReadMetaDataValue(i,output,delims):
	value = ""
	while i<len(output):
		ch = output[i]
		if ch in delims: break
		value = value + ch
		i = i + 1
	return i,value

def FCListMetaData(lfn=None):
	command = "FClistMetaData "
	if lfn:
		command = command + "-l '" + str(lfn) + "' "
	else:
		raise "No lookup key specified for FCListMetaData."

	rc,output = GetCommandOutput_NoStderr(command)
	if rc:
		if output: sys.stderr.write(output + "\n")
		raise "FCListMetaData failed: (status=" + str(rc) + ") " + command

	#Output is (key,value),(key,value),...
	#Return a dictionary.
	params = {}

	i = 0
	stripped_output = ""
	for line in string.split(output,"\n"):
		line = string.strip(line)
		if not line: continue
		if line[0] <> "(":
			#This gets rid of all the XML warnings and "errors"
			#that are reported in stdout.
			continue
		if stripped_output:
			stripped_output = stripped_output + "\n"
		stripped_output = stripped_output + line
	output = stripped_output
	while i<len(output):

		if output[i] != "(":
			raise "Cannot parse pool metadata at " + str(i) + " (expecting '('): " + output

		i = i + 1
		i,key = ReadMetaDataValue(i,output,",)")
		if i>=len(output) or output[i] != ",":
			raise "Cannot parse pool metadata at " + str(i) + " (expecting ','): " + output
		i = i + 1
		i,value = ReadMetaDataValue(i,output,")")
		if i>=len(output) or output[i] != ")":
			raise "Cannot parse pool metadata at " + str(i) + " (expecting ')'): " + output
		i = i + 1

		params[key] = value

		if i>=len(output): break
		if output[i] != ",":
			raise "Cannot parse pool metadata at " + str(i) + " (expecting ','): " + output
		i = i + 1

	return params

def PublishEVDFiles(tmp_evd_catalog,pool_file_catalog):
	command = "FCpublish -q DataType=EVD -u file:" + tmp_evd_catalog + \
			  " -d " + pool_file_catalog
	rc = os.system(command)
	if rc:
		sys.stderr.write("The following command failed:\n" \
						 + command + "\n")
		sys.exit(1)

def SummarizeOrcaErrors(log_file):
	sys.stdout.flush()
	print "########### Summary of error messages ###########"
	sys.stdout.flush()
	os.system("egrep -a '(CARF-ERROR)|(level\\[Error\\])' " + log_file + " | grep -a -v XMLDOMErrorMessanger")
	sys.stdout.flush()


def AttachRuns(runid_files,pool_file_catalog,owner,dataset):
	for fname in runid_files:
		log_file = "AttachRun." + str(os.getpid())
		command = "env POOL_CATALOG=" + pool_file_catalog + " " \
				+ "AttachRun -c " + fname + " " + owner + " " + dataset \
				+ " >& " + log_file
		print "Attaching " + fname
		rc = os.system(command)
		os.system("cat " + log_file)
		if rc:
			SummarizeOrcaErrors(log_file)
			sys.stderr.write("The following command failed: " + \
							 str(command) + "\n")
			sys.exit(1)
		os.remove(log_file)

def FixCollection(pool_file_catalog,owner,dataset):
	log_file = "FixColl." + str(os.getpid())
	command = "env POOL_CATALOG=" + pool_file_catalog + " " \
			+ "FixColl " + owner + " " + dataset + " >& " + log_file
	rc = os.system(command)
	os.system("cat " + log_file)
	if rc:
		SummarizeOrcaErrors(log_file)
		sys.stderr.write("The following command failed: " + \
						 str(command) + "\n")
		sys.exit(1)
	os.remove(log_file)

def InitGeometryPath():
	#zebra has problems with paths containing upper-case letters,
	#so symlink the geometry file to a local path.
	detinput = os.environ.get("DETINPUT")
	if not detinput: return
	detinput_basename = os.path.basename(detinput)
	if not os.path.exists(detinput_basename):
		os.symlink(detinput,detinput_basename)
	os.environ["DETINPUT"] = "./" + detinput_basename

def ReportEventsProcessed(log_file):
	command = """
awk '
  BEGIN {
	EVENTS=0
  }
  #For OSCAR and ORCA 7
  /Committing at event [0-9]+/ {
	EVENTS=$4
  }
  #For ORCA 8
  /Number of events analysed: +[0-9]+/ {
    match($0,/Number of events analysed: +([0-9]+)/,a)
	if(int(a[1])>int(EVENTS)) EVENTS=a[1]
  }
  #For ORCA 8 writeDST
  /[0-9]+ events analysed in total/ {
    match($0,/([0-9]+) events analysed in total/,a)
    if(int(a[1])>int(EVENTS)) EVENTS=a[1]
  }
  END {
	print \"WORK_DONE=\" EVENTS >> \"jug_monitor.vars\"
  }' """ + str(log_file)
	rc = os.system(command)
	if rc:
		print "Failed to report number of events processed."
		sys.exit(1)

def RunOrca(app,cards,pool_catalog,log_file):
	existing_log_length = 0
	if os.path.exists(log_file):
		F = file(log_file,"r")
		while F.readline():
			existing_log_length = existing_log_length + 1
		F.close()

	command = ""
	if pool_catalog:
		command = command + "export POOL_CATALOG=" + str(pool_catalog) + "\n"
	command = command + "exec " + app + " -c " + cards + " >> " + log_file + " 2>&1"
	rc = os.system(command)

	#Sometimes ORCA crashes _after_ successfully finishing the job.
	#In this case, we replace the process exit status with the exit
	#status of CARF Main.

	c_rc,carf_exit_status = GetCommandOutput("tail +" + str(1+existing_log_length) + " " + log_file + " | grep -a 'CARF Main ends with return code ' | tail -1 | awk '{ print $NF }'")

	#ORCA 8.2.0 does not appear to generate 'CARF Main ...' message.
	ignore_carf_exit_status = 0
	if c_rc or carf_exit_status == "":
		o_rc,orca_version = GetCommandOutput("tail +" + str(1+existing_log_length) + " " + log_file + " | grep -a -q 'ORCA_7'")
		if o_rc:
			#We are not running ORCA 7 and we did not see
			#'CARF Main ...' message, so ignore carf exit status.
			ignore_carf_exit_status = 1

	if not ignore_carf_exit_status:
		try:
			if c_rc:
				raise "grep failed"
			carf_exit_status = int(carf_exit_status)
		except:
			os.system("cat " + log_file)
			print "Failed to find CARF exit status."
			sys.exit(1)

		if carf_exit_status == 0:
			if rc <> 0:
				print "Fudging exit status: " + str(rc) + " --> " + str(carf_exit_status)
				rc = carf_exit_status

	if rc:
		os.system("cat " + log_file)
		SummarizeOrcaErrors(log_file)
		print app + " exited with status " + str(rc)
		sys.exit(1)
